#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
This is the top-level namespace of the groot_rocker_extensions package.
"""
##############################################################################
# Imports
##############################################################################

from . import bind                     # noqa
from . import colcon                   # noqa
from . import development_environment  # noqa
from . import git                      # noqa
from . import main                     # noqa
from . import named_prompt             # noqa
from . import nvidia                   # noqa
from . import snorriheim               # noqa
from . import ssh                      # noqa
from . import user                     # noqa
from . import work_directory           # noqa



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import getpass
import io
import os
import pathlib
import pwd

import docker

import groot_rocker
import groot_rocker.console as console

from . import utilities

##############################################################################
# Methods
##############################################################################


def remove_image(name: str):
    docker_client = groot_rocker.core.get_docker_client()
    docker_client.remove_image(image=name)

##############################################################################
# Tests
##############################################################################


class Nvidia(utilities.ExtensionTestCase):
    """
    By default, this test cleans up images and containers so that the user
    does not have to deal with dangling images (inadvertantly that also
    means that it's not caching between runs, but will make use of cached
    images in a single run).

    To debug via images, comment out the relevant explicit docker_client.remove_image
    command in remove_image() or instead, comment them individually below.
    """

    @classmethod
    def setUpClass(self):
        client = groot_rocker.core.get_docker_client()
        self.dockerfile_tags = []
        for distro_version in ['18.04']:
            dockerfile = (
                f"FROM ubuntu:{distro_version}\n"
                "RUN apt-get update && apt-get install glmark2 -y && apt-get clean\n"
                "CMD glmark2 --validate\n"
            )
            dockerfile_tag = "groot:" + f"test_nvidia_{distro_version}"
            iof = io.BytesIO((dockerfile % locals()).encode())
            image_id = client.build(fileobj=iof, tag=dockerfile_tag, forcerm=True)
            for unused_e in image_id:
                pass
            self.dockerfile_tags.append(dockerfile_tag)

    @classmethod
    def tearDownClass(self):
        # For quick debugging, comment out this method
        for image_name in self.dockerfile_tags:
            print(f"Image Name: {image_name}")
            # If the os detector images were made
            print(f"Teardown {image_name}")
            remove_image(name=image_name)
        try:
            remove_image("groot:os_detect_builder")
        except docker.errors.ImageNotFound:
            # some test methods do not build the os_detect_builder, that's ok
            pass

    def test_glmark2_validate_nvidia_not_enabled(self):
        for tag in self.dockerfile_tags:
            docker_tag = "groot:test_no_glmark2"
            dig = groot_rocker.core.DockerImageGenerator(
                active_extensions=[],
                cliargs={},
                base_image=tag
            )
            result = dig.build(image_name=docker_tag)
            utilities.assert_details(text="No Nvidia GLMark2 Build Result", expected=0, result=result)
            self.assertEqual(result, 0)
            result = dig.run()
            utilities.assert_details(text="No Nvidia GLMark2 Run Result", expected=1, result=result)
            self.assertEqual(result, 1)
            remove_image(name=docker_tag)

    def test_glmark2_validate_nvidia_enabled(self):
        extensions = groot_rocker.core.list_plugins()
        desired_extensions = ['nvidia', 'user']
        active_extensions = [e() for e in extensions.values() if e.get_name() in desired_extensions]
        for tag in self.dockerfile_tags:
            docker_tag = "groot:test_glmark2"
            dig = groot_rocker.core.DockerImageGenerator(
                active_extensions=active_extensions,
                cliargs={},
                base_image=tag
            )
            result = dig.build(image_name=docker_tag)
            utilities.assert_details(text="Nvidia GLMark2 Build Result", expected=0, result=result)
            self.assertEqual(result, 0)
            result = dig.run()
            utilities.assert_details(text="Nvidia GLMark2 Run Result", expected=0, result=result)
            self.assertEqual(result, 0)
            remove_image(name=docker_tag)

    def test_base_image_does_not_exist(self):
        mock_cli_args = {'base_image': 'ros:does-not-exist'}
        print(console.green + f"Checking base image '{mock_cli_args['base_image']}'" + console.reset)
        with self.assertRaises(SystemExit) as cm:
            self.extension.get_environment_subs(mock_cli_args)
        self.assertEqual(cm.exception.code, 1)
        utilities.assert_details(text="Base image does not exist", expected="SystemExit", result="SystemExit")

    def test_unsupported_base_image_version(self):
        mock_cli_args = {'base_image': 'ubuntu:17.04'}
        print(console.green + f"Checking base image '{mock_cli_args['base_image']}'" + console.reset)
        with self.assertRaises(SystemExit) as cm:
            self.extension.get_environment_subs(mock_cli_args)
        self.assertEqual(cm.exception.code, 1)
        utilities.assert_details(text="Unsupported Version", expected="SystemExit", result="SystemExit")
        remove_image("groot:os_detect_" + mock_cli_args['base_image'].replace(":", "_"))

    def test_unsupported_base_image_os(self):
        mock_cli_args = {'base_image': 'fedora'}
        print(console.green + f"Checking base image '{mock_cli_args['base_image']}'" + console.reset)
        with self.assertRaises(SystemExit) as cm:
            self.extension.get_environment_subs(mock_cli_args)
        self.assertEqual(cm.exception.code, 1)
        utilities.assert_details(text="Unsupported OS", expected="SystemExit", result="SystemExit")
        remove_image("groot:os_detect_" + mock_cli_args['base_image'].replace(":", "_"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import argparse
import em
import re
import unittest

import groot_rocker
import groot_rocker.console as console

##############################################################################
# Helper Methods
##############################################################################


def assert_details(text, expected, result):
    print(console.green + text +
          "." * (40 - len(text)) +
          console.cyan + "{}".format(expected) +
          console.yellow + " [{}]".format(result) +
          console.reset)


##############################################################################
# Imports
##############################################################################


class ExtensionTestCase(unittest.TestCase):

    def setUp(self):
        """
        Children should implement their own setup() method and:

         * set self.extension_name appropriately
         * call this (the parent setup() method)
        """
        self.extension_name = re.sub(r'(?<!^)(?=[A-Z])', '_', self.__class__.__name__).lower()
        # Work around interference between empy Interpreter
        # stdout proxy and test runner. empy installs a proxy on stdout
        # to be able to capture the information.
        # And the test runner creates a new stdout object for each test.
        # This breaks empy as it assumes that the proxy has persistent
        # between instances of the Interpreter class
        # empy will error with the exception
        # "em.Error: interpreter stdout proxy lost"
        em.Interpreter._wasProxyInstalled = False
        # Ensure that a user of this class has set self.extension_name
        try:
            hasattr(self, "extension_name")
        except AttributeError:
            raise AssertionError("Children must provide a value for self.extension_name (e.g. 'user')")
        assert isinstance(self.extension_name, str), "self.extension_name must be a str type"
        self.extensions = groot_rocker.core.list_plugins()
        self.extension_class = self.extensions[self.extension_name]
        self.extension = self.extension_class()
        print("\n")  # clearly delineates between test methods

    def test_extension_name(self):
        assert_details(
            text="Extension Name",
            expected=self.extension_name,
            result=self.extension.get_name()
        )
        self.assertEqual(self.extension.get_name(), self.extension_name)

    def test_load_parser_correctly(self):
        """
        A helper function to test that the plugins at least
        register an option for their own name.
        """
        parser = argparse.ArgumentParser(description='test_parser')
        self.extension.register_arguments(parser)
        argument_name = groot_rocker.extensions.name_to_argument(self.extension.get_name())
        found_string = False
        for action in parser._actions:
            option_strings = getattr(action, 'option_strings', [])
            if argument_name in option_strings:
                found_string = True
                break
        assert_details(
            text="Extension Name as Argparse Arg",
            expected=True,
            result=found_string
        )
        self.assertTrue(found_string)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import getpass
import os
import pathlib
import pwd

from . import utilities

##############################################################################
# Tests
##############################################################################


class User(utilities.ExtensionTestCase):

    def test_environment_queries(self):
        env_subs = self.extension.get_environment_subs()
        utilities.assert_details(
            text="Environment::gid",
            expected=os.getgid(),
            result=env_subs['gid']
        )
        self.assertEqual(env_subs['gid'], os.getgid())
        utilities.assert_details(
            text="Environment::uid",
            expected=os.getuid(),
            result=env_subs['uid']
        )
        self.assertEqual(env_subs['uid'], os.getuid())
        utilities.assert_details(
            text="Environment::name",
            expected=getpass.getuser(),
            result=env_subs['name']
        )
        self.assertEqual(env_subs['name'],  getpass.getuser())
        utilities.assert_details(
            text="Environment::dir",
            expected=str(pathlib.Path.home()),
            result=env_subs['dir']
        )
        self.assertEqual(env_subs['dir'],  str(pathlib.Path.home()))
        utilities.assert_details(
            text="Environment::gecos",
            expected=pwd.getpwuid(os.getuid()).pw_gecos,
            result=env_subs['gecos']
        )
        self.assertEqual(env_subs['gecos'],  pwd.getpwuid(os.getuid()).pw_gecos)
        utilities.assert_details(
            text="Environment::shell",
            expected=pwd.getpwuid(os.getuid()).pw_shell,
            result=env_subs['shell']
        )
        self.assertEqual(env_subs['shell'],  pwd.getpwuid(os.getuid()).pw_shell)

        mock_cliargs = {}
        snippet = self.extension.get_snippet(mock_cliargs).splitlines()

        uid_line = [l for l in snippet if '--uid' in l][0]
        self.assertTrue(str(os.getuid()) in uid_line)

        self.assertEqual(self.extension.get_preamble(mock_cliargs), '')
        self.assertEqual(self.extension.get_docker_args(mock_cliargs), '')

        self.assertTrue('mkhomedir_helper' in self.extension.get_snippet(mock_cliargs))
        home_active_cliargs = mock_cliargs
        home_active_cliargs['home'] = True
        self.assertFalse('mkhomedir_helper' in self.extension.get_snippet(home_active_cliargs))




#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Switch to the specified working directory on entry into the container.
"""

##############################################################################
# Imports
##############################################################################

import em
import os
import pkgutil
import pwd
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class User(groot_rocker.extensions.RockerExtension):
    """
    Map the currently executing user to the docker environment. This essentially
    just recreates user and group information in the image. It also specifies
    the WORKDIR as the home directory if not otherwise set by the WorkDirectory
    extension.
    """

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def get_environment_subs(self) -> typing.Dict[str, str]:
        user_vars = ['name', 'uid', 'gid', 'gecos', 'dir', 'shell']
        userinfo = pwd.getpwuid(os.getuid())
        return {
            k: getattr(userinfo, 'pw_' + k)
            for k in user_vars
        }

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        """
        """
        snippet = pkgutil.get_data('groot_rocker_extensions', f"templates/{User.get_name()}.Dockerfile.em").decode('utf-8')
        substitutions = self.get_environment_subs()
        substitutions['home_extension_active'] = True if 'home' in cli_args and cli_args['home'] else False
        substitutions['work_directory_active'] = True if 'work_directory' in cli_args and cli_args['work_directory'] else False
        return em.expand(snippet, substitutions)

    def get_docker_args(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--user',
            action='store_true',
            default=defaults.get('user', None),
            help="mount the current user's id and run as that user"
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Enable nvidia in the docker image/container.
"""

##############################################################################
# Imports
##############################################################################

import em
import pkgutil
import re
import sys
import typing

import groot_rocker
import groot_rocker.console as console

##############################################################################
# Extension
##############################################################################


class Nvidia(groot_rocker.extensions.RockerExtension):
    """
    This class pulls the nvidia docker images and incorporates it on top
    of an existing base image.
    """
    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def __init__(self):
        self.supported_distros = ['Ubuntu']  # 'Debian GNU/Linux'
        self.supported_versions = ['18.04', '20.04']  # '10'

    def get_environment_subs(self, cli_args: typing.Dict[str, str] = {}) -> typing.Dict[str, str]:
        substitutions = {}

        # non static elements test every time
        detected_os = groot_rocker.os_detector.detect_os(
            cli_args['base_image'],
            print,
            nocache=cli_args.get('nocache', False)
        )
        if detected_os is None:
            console.error(f"unable to detect os for base image '{cli_args['base_image']}', maybe the base image does not exist")
            sys.exit(1)
        dist, ver, unused_codename = detected_os

        substitutions['image_distro_id'] = dist
        if substitutions['image_distro_id'] not in self.supported_distros:
            console.error(f"distro id '{substitutions['image_distro_id']}' not in supported list {self.supported_distros}")
            sys.exit(1)
        else:
            # for embedding in the nvidia_preamble snippet to identify the nvidia opengl image
            substitutions['image_distro_id'] = substitutions['image_distro_id'].lower()
        substitutions['image_distro_version'] = ver
        if substitutions['image_distro_version'] not in self.supported_versions:
            console.error(f"distro '{dist}', version '{ver}' not in supported list {self.supported_versions}")
            sys.exit(1)
            # TODO(tfoote) add a standard mechanism for checking preconditions and disabling plugins
        return substitutions

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, cli_args: typing.Dict[str, str]) -> str:
        preamble = pkgutil.get_data('groot_rocker_extensions', f"templates/{Nvidia.get_name()}_preamble.Dockerfile.em").decode('utf-8')
        substitutions = self.get_environment_subs(cli_args)
        return em.expand(preamble, substitutions)

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        snippet = pkgutil.get_data('groot_rocker_extensions', f"templates/{Nvidia.get_name()}_snippet.Dockerfile.em").decode('utf-8')
        substitutions = self.get_environment_subs(cli_args)
        return em.expand(snippet, substitutions)

    def get_docker_args(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return " --gpus all --volume /tmp/.X11-unix:/tmp/.X11-unix:ro --env=DISPLAY"

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--nvidia',
            action='store_true',
            default=defaults.get('nvidia', None),
            help="enable nvidia"
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Configure and install tools for a generic linux development environment.
"""

##############################################################################
# Imports
##############################################################################

import em
import os
import pkgutil
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class DevelopmentEnvironment(groot_rocker.extensions.RockerExtension):
    """
    Installs various system dependencies and configures environment variables,
    locales, suitable for a generic linux development environment.
    """

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        """
        """
        snippet = pkgutil.get_data(
            'groot_rocker_extensions',
            f"templates/{DevelopmentEnvironment.get_name()}.Dockerfile.em").decode('utf-8')
        substitutions = {
            "default_timezone": os.readlink("/etc/localtime"),
            "system_dependencies": (
               "apt-utils bash-completion build-essential curl debian-keyring "
               "debian-archive-keyring git iproute2 iputils-ping locales lsb-release "
               "net-tools openssh-client openssh-server python3-dev software-properties-common "
               "sudo vim wget xdot x11-apps "
            ),
            "language": "en_US.UTF-8",
            # Richest color set - https://www.gnu.org/software/gettext/manual/html_node/The-TERM-variable.html
            "terminal": "xterm-256color",
        }
        return em.expand(snippet, substitutions)

    def get_docker_args(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--development-environment', '--devenv',
            action='store_true',
            default=defaults.get('development_environment', None),
            help="setup a generic development environment"
        )



#!/usr/bin/env python3

##############################################################################
# Imports
##############################################################################

import setuptools

##############################################################################
# Requirements
##############################################################################

install_requires = [
    # build
    'setuptools',
    # runtime
    'distro',
    'groot_rocker',
]

extras_require = {
    'tests': ['codecov', 'coverage', 'nose', 'pytest'],
    'packaging': ['stdeb', 'twine']
}

##############################################################################
# Setup
##############################################################################

d = setuptools.setup(
    name='groot_rocker_extensions',
    version='0.2.4',
    packages=setuptools.find_packages(exclude=['tests*', 'docs*']),
    package_data={'groot_rocker_extensions': ['templates/*.em']},
    install_requires=install_requires,
    extras_require=extras_require,
    author='Daniel Stonier',
    maintainer='Daniel Stonier <d.stonier@gmail.com>',
    url='http://github.com/stonier/groot_rocker_extensions',
    keywords=['Docker'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries'
    ],
    description="groot_rocker extensions for groot workflows",
    long_description="A few groot_rocker extensiosn for groot workflows.",
    license='BSD',
    python_requires='>=3.0',
    entry_points={
        'console_scripts': [
            'groot-rocker-workspace = groot_rocker_extensions.main:main_workspace',
        ],
        'groot_rocker.extensions': [
            'bind = groot_rocker_extensions.bind:Bind',
            'colcon = groot_rocker_extensions.colcon:Colcon',
            'development_environment = groot_rocker_extensions.development_environment:DevelopmentEnvironment',
            'git = groot_rocker_extensions.git:Git',
            'named_prompt = groot_rocker_extensions.named_prompt:NamedPrompt',
            'nvidia = groot_rocker_extensions.nvidia:Nvidia',
            'ssh = groot_rocker_extensions.ssh:Ssh',
            'user = groot_rocker_extensions.user:User',
            'work_directory = groot_rocker_extensions.work_directory:WorkDirectory',
        ],
    },
)
